﻿/*
  COPYRIGHT 2022 - PROPERTY OF TOBII AB
  -------------------------------------
  2022 TOBII AB - KARLSROVAGEN 2D, DANDERYD 182 53, SWEDEN - All Rights Reserved.

  NOTICE:  All information contained herein is, and remains, the property of Tobii AB and its suppliers, if any.
  The intellectual and technical concepts contained herein are proprietary to Tobii AB and its suppliers and may be
  covered by U.S.and Foreign Patents, patent applications, and are protected by trade secret or copyright law.
  Dissemination of this information or reproduction of this material is strictly forbidden unless prior written
  permission is obtained from Tobii AB.
*/
using System;
using System.Collections.Generic;
using Tobii.StreamEngine;

namespace Tobii.StreamEngine
{
    /// <summary>
    /// Interface for a wrapper of communication with Tobii Stream Engine.
    /// This interface can be used to mock stream engine functionality in unit tests.
    /// </summary>
    public interface IStreamEngineWrapper
    {
        /// <summary>
        /// Start calibration for a device
        /// </summary>
        /// <param name="device">Pointer to device context to start calibration for</param>
        /// <param name="enabledEye">What eyes are enabled for this calibration (left/right/both)</param>
        /// <returns>The result of the attempt to start calibration</returns>
        tobii_error_t CalibrationStart(IntPtr device, tobii_enabled_eye_t enabledEye);

        /// <summary>
        /// Stop calibration for a device
        /// </summary>
        /// <param name="device">Pointer to device context to stop calibration for</param>
        /// <returns>The result of the attempt to stop calibration</returns>
        tobii_error_t CalibrationStop(IntPtr device);

        /// <summary>
        /// Subscribe to receive wearable gaze data with a consumer callback
        /// </summary>
        /// <param name="device">Pointer to device context to subscribe to</param>
        /// <param name="callback">Callback to call with gaze data when callbacks are processed</param>
        /// <param name="userData">Pointer to user data to be used in the callback</param>
        /// <returns>The result of the attempt to subscribe</returns>
        tobii_error_t WearableConsumerDataSubscribe(IntPtr device, tobii_wearable_consumer_data_callback_t callback, IntPtr userData = default);

        /// <summary>
        /// Request device to process callbacks. If your application subscribed to receive wearable gaze data, the registered callback will be called.
        /// </summary>
        /// <param name="device">Pointer to device context to process callbacks for</param>
        /// <returns>The result of the attempt to process callbacks</returns>
        tobii_error_t DeviceProcessCallbacks(IntPtr device);

        /// <summary>
        /// Compute eye models and immediately apply them
        /// </summary>
        /// <param name="device">Pointer to device context to request compute and apply</param>
        /// <returns>The result of the attempt to compute and apply</returns>
        tobii_error_t CalibrationComputeAndApply(IntPtr device);

        /// <summary>
        /// Collect calibration data in specified 3d position
        /// </summary>
        /// <param name="device">Pointer to device context to request collection</param>
        /// <param name="x">X coordinate for the position</param>
        /// <param name="y">Y coordinate for the position</param>
        /// <param name="z">Z coordinate for the position</param>
        /// <returns>The result of the attempt to collect calibration data</returns>
        tobii_error_t CalibrationCollectData3d(IntPtr device, float x, float y, float z);

        /// <summary>
        /// Unsubscribe from receiving wearable gaze data
        /// </summary>
        /// <param name="device">Pointer to device context to unsubscribe from</param>
        /// <returns>The result of the attempt to unsubscribe</returns>
        tobii_error_t WearableConsumerDataUnsubscribe(IntPtr device);

        /// <summary>
        /// Create device context without stream engine license
        /// </summary>
        /// <param name="api">Pointer to stream engine api to use</param>
        /// <param name="url">Url of the device to create</param>
        /// <param name="fieldOfUse">Value indicating if the application will store or transfer eye tracking data
        /// <list type="bullet">
        /// <item>
        ///     <term>TOBII_FIELD_OF_USE_STORE_OR_TRANSFER_TRUE</term>
        ///     <description>Should be used if the application will store eye tracking data on the local machine, a server, to another application, or a cloud system</description>
        /// </item>
        /// <item>
        ///     <term>TOBII_FIELD_OF_USE_STORE_OR_TRANSFER_FALSE</term>
        ///     <description>Should be used if the application will not store eye tracking data</description>
        /// </item>
        /// </list>
        /// </param>
        /// <param name="device">Pointer to created device context</param>
        /// <returns>The result of the attempt to create device context</returns>
        tobii_error_t DeviceCreate(IntPtr api, string url, Interop.tobii_field_of_use_t fieldOfUse, out IntPtr device);

        /// <summary>
        /// Create device context using stream engine license keys
        /// </summary>
        /// <param name="api">Pointer to stream engine api to use</param>
        /// <param name="url">Url of the device to create</param>
        /// <param name="fieldOfUse">Value indicating if the application will store or transfer eye tracking data
        /// <list type="bullet">
        /// <item>
        ///     <term>TOBII_FIELD_OF_USE_STORE_OR_TRANSFER_TRUE</term>
        ///     <description>Should be used if the application will store eye tracking data on the local machine, a server, to another application, or a cloud system</description>
        /// </item>
        /// <item>
        ///     <term>TOBII_FIELD_OF_USE_STORE_OR_TRANSFER_FALSE</term>
        ///     <description>Should be used if the application will not store eye tracking data</description>
        /// </item>
        /// </list>
        /// </param>
        /// <param name="licenseKeys">An array of license keys</param>
        /// <param name="licenseResults">Results of license validation. Each item in the result list corresponds to a license key provided in licenseKeys parameter</param>
        /// <param name="device">Pointer to created device context</param>
        /// <returns>The result of the attempt to create device context</returns>
        tobii_error_t DeviceCreateEx(IntPtr api, string url, Interop.tobii_field_of_use_t fieldOfUse, string[] licenseKeys, List<tobii_license_validation_result_t> licenseResults, out IntPtr device);

        /// <summary>
        /// Destroy the device context instance and release related resources
        /// </summary>
        /// <param name="device">Pointer to device context to destroy</param>
        /// <returns>The result of the attempt to destroy device context</returns>
        tobii_error_t DeviceDestroy(IntPtr device);

        /// <summary>
        /// Create stream engine api context
        /// </summary>
        /// <param name="api">Pointer to created api context</param>
        /// <param name="customLog">Custom function to handle log output. See documentation for more details</param>
        /// <returns>The result of the attempt to create api context</returns>
        tobii_error_t ApiCreate(out IntPtr api, tobii_custom_log_t customLog);

        /// <summary>
        /// Destroy the api context instance and release related resources
        /// </summary>
        /// <param name="api">Stream engine api context to destroy</param>
        /// <returns>The result of the attempt to destroy api context</returns>
        tobii_error_t ApiDestroy(IntPtr api);

        /// <summary>
        /// Enumerate local device urls and call receiver function with device url and provided pointer to user data
        /// </summary>
        /// <param name="api">Pointer to stream engine api to use</param>
        /// <param name="receiverFunction">Callback to call with device url</param>
        /// <param name="userData">Pointer to user data to be used in callback</param>
        /// <returns>The result of the attempt to enumerate device urls</returns>
        tobii_error_t EnumerateLocalDeviceUrlsInternal(IntPtr api, tobii_device_url_receiver_t receiverFunction, IntPtr userData);

        /// <summary>
        /// Retrieve information about a device 
        /// </summary>
        /// <param name="device">Pointer to device context to get information for</param>
        /// <param name="deviceInfo">Information about the device</param>
        /// <returns>The result of the attempt to get information</returns>
        tobii_error_t GetDeviceInfo(IntPtr device, out tobii_device_info_t deviceInfo);

        /// <summary>
        /// Establish a new connection after a disconnect
        /// </summary>
        /// <param name="device">Pointer to device context to reconnect to</param>
        /// <remarks>When receiving the error code TOBII_ERROR_CONNECTION_FAILED, it is necessary to explicitly request reconnection</remarks>
        /// <returns>The result of the attempt to reconnect</returns>
        tobii_error_t DeviceReconnect(IntPtr device);
    }
}